open Ast

module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

let string_of_program (vars, funcs) =
  (* Put function declarations in a symbol table *)
  let func_decls = List.fold_left
        (fun funcs fdecl -> 
            if NameMap.mem fdecl.fname funcs then
                raise (Failure ("Function: '" ^ fdecl.fname ^ 
                    "' has already been defined."))
            else            
                NameMap.add fdecl.fname fdecl funcs) NameMap.empty funcs
    in
    let rec string_of_fdecl globals fdecl actuals =
    
    let rec string_of_expr globals locals fname = function
        Literal(l) -> { varname = string_of_int l;vartype = "int" }
      | Floatlit(f) -> { varname = string_of_float f;vartype = "float" }
      | String(s) -> { varname = s;vartype = "string" }
      | Bool(s) -> { varname = s;vartype = "boolean" }      
      | Id(s) ->
            if NameMap.mem s locals then 
    	        NameMap.find s locals
            else if NameMap.mem s globals then 
    	        NameMap.find s globals
    	    else raise (Failure ("Undeclared identifier found in expression: '" 
    	        ^ s ^ "' in function: '" ^ fname ^ "'"))
      | Datatype(t) -> 
            if t = "parseerror" then
                raise ((Failure ("Syntax error")))
            else { varname = "null" ;vartype = t }
      | Uminus(e) -> let v = string_of_expr globals locals fname e in 
            if v.vartype = "int" || v.vartype = "float" then
                v
            else
                raise (Failure ("unary - operator can only be applied to int or float expressions" 
                                ^ ", but here used with '" ^ v.varname ^ "', type: " 
                                ^ v.vartype ^ " in function: '" ^ fname ^ "'"))
      | Not(e) -> let v = string_of_expr globals locals fname e in
            if v.vartype = "boolean" then
                v
            else
                raise (Failure ("not operator can only be applied to boolean expression" 
                                ^ ", but here used with '" ^ v.varname ^ "', type: " 
                                ^ v.vartype ^ " in function: '" ^ fname ^ "'"))
      | Binop(e1, o, e2) ->
          let v1 = string_of_expr globals locals fname e1 in
          (match o with
             Add ->
             let v2 = string_of_expr globals locals fname e2 in
               if v1.vartype = "string" then
                 if v2.vartype = "image" || v2.vartype = "kernel" then
                   raise (Failure ("Cannot concatenate " ^ v2.vartype ^ " type with string type " ^
                                   "in function: '" ^ fname ^ "'"))
                 else v1
               else if v2.vartype = "string" then
                 if v1.vartype = "image" || v1.vartype = "kernel" then
                   raise (Failure ("Cannot concatenate " ^ v1.vartype ^ " type with string type " ^
                                   "in function: '" ^ fname ^ "'"))
                 else v2
               else if v1.vartype = "image" && v2.vartype = "int" then
                      v1
               (*else if v2.vartype = "image" && v1.vartype = "int" then
                      v2*)
               else if v1.vartype = "kernel" && v2.vartype = "float" then
                      v1
               else if v1.vartype = "boolean" || v2.vartype = "boolean" then
                 raise (Failure ("+ operator cannot be applied to boolean operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
               else if v1.vartype <> v2.vartype then
                 raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                  v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                  ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                  ^ "'"))
               else v1 
    	   | Sub -> 
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "string" || v2.vartype = "string" then
                  raise (Failure ("Operators '-' " ^
                                  "cannot be applied to a string, in function: '" ^ fname ^ "'"))
                else if v1.vartype = "image" && v2.vartype = "int" then
                      v1
                (*else if v2.vartype = "image" && v1.vartype = "int" then
                      v2*)
                else if v1.vartype = "kernel" && v2.vartype = "float" then
                      v1
               else if v1.vartype = "boolean" || v2.vartype = "boolean" then
                 raise (Failure ("- operator cannot be applied to boolean operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
                else if v1.vartype <> v2.vartype then 
                  raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                   v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                   ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                   ^ "'")) 
                else v1
           | Equal | Neq ->
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype <> v2.vartype then 
                  raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                   v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                   ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                   ^ "'")) 
                else { varname = "null"; vartype = "boolean" }
    	   | Mult | Div -> 
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "string" || v2.vartype = "string" then
                  raise (Failure ("Operators '*', '/' " ^
                                  "cannot be applied to a string, in function: '" ^ fname ^ "'"))
                else if v1.vartype = "image" && v2.vartype = "int" then
                    v1
                (*else if v2.vartype = "image" && v1.vartype = "int" then
                    v2*)
                else if v1.vartype = "kernel" && v2.vartype = "float" then
                    v1
                else if v1.vartype = "boolean" || v2.vartype = "boolean" then
                 raise (Failure ("/, * operators cannot be applied to boolean operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
                else if v1.vartype <> v2.vartype then 
                  raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                   v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                   ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                   ^ "'")) 
                else v1
    	   | Less | Leq | Greater | Geq -> 
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "string" || v2.vartype = "string" then
                  raise (Failure ("Operators '<', '<=', '>', '>=' " ^
                                  "cannot be applied to a string, in function: '" ^ fname ^ "'"))
               else if v1.vartype = "boolean" || v2.vartype = "boolean" then
                 raise (Failure ("<,<=,>,>= operators cannot be applied to boolean operands" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
                else if v1.vartype <> v2.vartype then 
                  raise (Failure ("Type mismatch in binary operation between: '" ^ 
                                   v1.varname ^ "', type: " ^ v1.vartype ^ " and '" ^ v2.varname
                                   ^ "', type: " ^ v2.vartype ^ " in function: '" ^ fname 
                                   ^ "'")) 
                else { varname = "null"; vartype = "boolean" }
    	   | And | Or -> 
             let v2 = string_of_expr globals locals fname e2 in
                if v1.vartype = "boolean" && v2.vartype = "boolean" then
                    v1
                else
                    raise (Failure ("and, or operators can only be applied to boolean expressions" 
                                  ^ ", but here used with '" ^ v1.varname ^ "', type: " 
                                  ^ v1.vartype ^ " and '" ^ v2.varname ^ "', type: " 
                                  ^ v2.vartype ^ " in function: '" ^ fname ^ "'"))
          | Convolve ->
            let v2 = string_of_expr globals locals fname e2 in
            if v1.vartype = "image" && v2.vartype = "kernel" then 
                v1
            else
	            raise (Failure ("@ operator needs image and kernel as operands,"
	            ^ " where image is the left operand, and kernel is the right operand,"
	            ^ " but here got '" ^ v1.vartype ^ ": " ^ v1.varname ^ "' and '" 
	            ^ v2.vartype ^ ": " ^ v2.varname ^ "' in function: '" ^ fname ^ "'"))
          | Power ->
            let v2 = string_of_expr globals locals fname e2 in
            if v1.vartype = "int" && v2.vartype = "int" then 
                v1
            else if v1.vartype = "float" && v2.vartype = "float" then 
                v1
            else
	            raise (Failure ("^ operator needs both operands as int or float,"
	            ^ " but here got '" ^ v1.vartype ^ ": " ^ v1.varname ^ "' and '" 
	            ^ v2.vartype ^ ": " ^ v2.varname ^ "' in function: '" ^ fname ^ "'")))
      | Assign(v, e) -> 
            if NameMap.mem v locals then
                let var1 = NameMap.find v locals in
                let var2 = string_of_expr globals locals fname e in
    	        if var1.vartype <> var2.vartype then 
    	            raise (Failure ("Type mismatch in assignment operation between: '" 
    	                ^ var1.varname ^ "', type: " ^ var1.vartype ^ " and '" 
    	                ^ var2.varname ^ "', type: " ^ var2.vartype 
    	                ^ " in function: '" ^ fname ^ "'")) 
    	        else var1
    	  else if NameMap.mem v globals then
    	        let var1 = NameMap.find v globals in
                let var2 = string_of_expr globals locals fname e in
    	        if var1.vartype <> var2.vartype then 
    	            raise (Failure ("Type mismatch in assignment operation between: '" 
    	                ^ var1.varname ^ "', type: " ^ var1.vartype ^ " and '" 
    	                ^ var2.varname ^ "', type: " ^ var2.vartype 
    	                ^ " in function: '" ^ fname ^ "'")) 
    	        else var1
    	  else raise (Failure ("Undeclared identifier found in assignment: '" 
    	        ^ v ^ "' in function: '" ^ fname ^ "'"))
      | AssignPixel(id, row, col, channel, value) ->  
            let varchannel = string_of_expr globals locals fname channel in
            if varchannel.vartype <> "string" then
                raise (Failure ("The channel in pixel assignment must be of type string" 
                    ^ ", but here used with: '" ^ varchannel.vartype ^ "' in function: '" 
                    ^ fname ^ "'"))
            else  
                let varvalue = string_of_expr globals locals fname value in  
                let varRow = string_of_expr globals locals fname row in  
                let varCol = string_of_expr globals locals fname col in
                if  varvalue.vartype <> "int" then  
                    raise (Failure ("Type mismatch in image pixel assignment."   
                        ^ " The value must be of type int but here used with type '"
                        ^ varvalue.vartype ^ "' in function: '" ^ fname ^ "'"))  
                else
                    if NameMap.mem id locals then  
                        let varId = NameMap.find id locals in  
                        if varId.vartype <> "image" then  
                            raise (Failure ("Expecting image type but got '" ^ varId.vartype ^  
                                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'"))   
                        else if varRow.vartype <> varCol.vartype || varRow.vartype <> "int" then   
                            raise (Failure ("Type mismatch in image access. Both row and column"   
                                ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))  
                        else { varname = "null";vartype = "int" }   
                    else if NameMap.mem id globals then  
                        let varId = NameMap.find id globals in  
                        if varId.vartype <> "image" then  
                        raise (Failure ("Expecting image type but got '" ^ varId.vartype ^  
                            "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'"))   
                        else if varRow.vartype <> varCol.vartype && varRow.vartype <> "int" then   
                            raise (Failure ("Type mismatch in image access. Both row and column"   
                                ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))  
                        else { varname = "null";vartype = "int" }
            	  else raise (Failure ("Undeclared identifier found in assignment: '" 
            	        ^ id ^ "' in function: '" ^ fname ^ "'"))
      | AssignKernel(id, row, col, value) ->  
            let varvalue = string_of_expr globals locals fname value in  
            let varRow = string_of_expr globals locals fname row in  
            let varCol = string_of_expr globals locals fname col in
            if  varvalue.vartype <> "float" then  
                raise (Failure ("Type mismatch in kernel value assignment."   
                    ^ " The value must be of type float but here used with type '"
                    ^ varvalue.vartype ^ "' in function: '" ^ fname ^ "'"))  
            else
                if NameMap.mem id locals then  
                    let varId = NameMap.find id locals in  
                    if varId.vartype <> "kernel" then  
                        raise (Failure ("Expecting kernel type but got '" ^ varId.vartype ^  
                            "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'"))   
                    else if varRow.vartype <> varCol.vartype || varRow.vartype <> "int" then   
                        raise (Failure ("Type mismatch in kernel access. Both row and column"   
                            ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))  
                    else { varname = "null";vartype = "float" }   
                else if NameMap.mem id globals then  
                    let varId = NameMap.find id globals in  
                    if varId.vartype <> "kernel" then  
                    raise (Failure ("Expecting kernel type but got '" ^ varId.vartype ^  
                        "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'"))   
                    else if varRow.vartype <> varCol.vartype && varRow.vartype <> "int" then   
                        raise (Failure ("Type mismatch in kernel access. Both row and column"   
                            ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))  
                    else { varname = "null";vartype = "float" }
        	  else raise (Failure ("Undeclared identifier found in assignment: '" 
        	        ^ id ^ "' in function: '" ^ fname ^ "'"))
      | ImageAccess(id, row, col, channel) ->
            let varchannel = string_of_expr globals locals fname channel in
            if varchannel.vartype <> "string" then
                raise (Failure ("The channel in pixel access must be of type string" ^
                      ", but here used with: '" ^ varchannel.vartype ^ "' in function: '" ^ fname ^ "'"))
    	    else
                let varRow = string_of_expr globals locals fname row in
                let varCol = string_of_expr globals locals fname col in
                if NameMap.mem id locals then
                    let varId = NameMap.find id locals in
                    if varId.vartype <> "image" then
        	            raise (Failure ("Expecting image type but got '" ^ varId.vartype ^
        	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
        	        else if varRow.vartype <> varCol.vartype || varRow.vartype <> "int" then 
        	            raise (Failure ("Type mismatch in image access. Both row and column" 
        	                ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))
        	        else { varname = "null";vartype = "int" } 
        	  else if NameMap.mem id globals then
                    let varId = NameMap.find id globals in
                    if varId.vartype <> "image" then
        	            raise (Failure ("Expecting image type but got '" ^ varId.vartype ^
        	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
        	        else if varRow.vartype <> varCol.vartype && varRow.vartype <> "int" then 
        	            raise (Failure ("Type mismatch in image access. Both row and column" 
        	                ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))
        	        else { varname = "null";vartype = "int" } 
        	  else raise (Failure ("Undeclared identifier found in function: '" ^ fname ^ "'"))      
      | KernelAccess(id, row, col) ->
            let varRow = string_of_expr globals locals fname row in
            let varCol = string_of_expr globals locals fname col in
            if NameMap.mem id locals then
                let varId = NameMap.find id locals in
                if varId.vartype <> "kernel" then
    	            raise (Failure ("Expecting kernel type but got '" ^ varId.vartype ^
    	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
    	        else if varRow.vartype <> varCol.vartype || varRow.vartype <> "int" then 
    	            raise (Failure ("Type mismatch in kernel access. Both row and column" 
    	                ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))
    	        else { varname = "null";vartype = "float" } 
    	    else if NameMap.mem id globals then
                let varId = NameMap.find id globals in
                if varId.vartype <> "kernel" then
    	            raise (Failure ("Expecting kernel type but got '" ^ varId.vartype ^
    	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
    	        else if varRow.vartype <> varCol.vartype && varRow.vartype <> "int" then 
    	            raise (Failure ("Type mismatch in kernel access. Both row and column" 
    	                ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))
    	        else { varname = "null";vartype = "float" } 
    	    else raise (Failure ("Undeclared identifier found in function: '" ^ fname ^ "'"))
      | Imnew(row, col, format) ->
            if format <> "rgb" && format <> "RGB" && format <> "gray" && format <> "GRAY" 
               && format <> "grey" && format <> "GREY" then
                raise (Failure ("The image format must be either rgb or gray scale" ^
                      ", but here used with: '" ^ format ^ "' in function: '" ^ fname ^ "'"))
    	    else
                let varRow = string_of_expr globals locals fname row in
                let varCol = string_of_expr globals locals fname col in
    	        if varRow.vartype <> varCol.vartype || varRow.vartype <> "int" then 
    	            raise (Failure ("Type mismatch in image access. Both row and column" 
    	                ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))
                else
    	            { varname = "imnew";vartype = "image" }
      | Kernelnew(row, col) ->
                let varRow = string_of_expr globals locals fname row in
                let varCol = string_of_expr globals locals fname col in
    	        if varRow.vartype <> varCol.vartype || varRow.vartype <> "int" then 
    	            raise (Failure ("Type mismatch in image access. Both row and column" 
    	                ^ " accessors must be of type int; in function: '" ^ fname ^ "'"))
                else
    	            { varname = "kernelnew";vartype = "kernel" }
      | Call("print", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'print'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype <> "image" && param.vartype <> "kernel" then 
                            { varname = "print" ; vartype = "void" }
                        else
                            raise (Failure ("Argument passed to function 'print' cannot be " 
                                ^ "image or kernel type, called from function: '" ^ fname ^ "'"))
      | Call("tostring", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'tostring'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype <> "image" && param.vartype <> "kernel" then 
                            { varname = "tostring" ; vartype = "string" }
                        else
                            raise (Failure ("Argument passed to function 'tostring' cannot be " 
                                ^ "image or kernel type, called from function: '" ^ fname ^ "'"))
      | Call("toint", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("function 'toint' takes just one parameter" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "float" || param.vartype = "string" then 
                            { varname = "toint" ; vartype = "int" }
                        else
                            raise (Failure ("Argument passed to function 'toint' can only be " 
                                ^ "either float or string type, called from function: '" ^ fname ^ "'"))
      | Call("tofloat", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("function 'tofloat' takes just one parameter" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "int" || param.vartype = "string" then 
                            { varname = "tofloat" ; vartype = "float" }
                        else
                            raise (Failure ("Argument passed to function 'tofloat' can only be " 
                                ^ "either int or string type, called from function: '" ^ fname ^ "'"))
      | Call("togray", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("function 'togray' takes just one parameter" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "image"then 
                            { varname = "togray" ; vartype = "image" }
                        else
                            raise (Failure ("Argument passed to function 'togray' can only be " 
                                ^ "image type, called from function: '" ^ fname ^ "'"))
      | Call("imread", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'imread'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "string" then 
                            { varname = "imread" ; vartype = "image" }
                        else
                            raise (Failure ("Argument passed to function 'imread' must be of string type" 
                                ^ ", called from function: '" ^ fname ^ "'"))
      | Call("imsave", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 2 then 
                    raise (Failure ("Wrong number of arguments passed to function 'imsave'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param1 = List.nth actuals 1 in
                        if param1.vartype = "image" then
                            let param2 = List.nth actuals 0 in
                            if param2.vartype = "string" then
                                { varname = "imsave" ; vartype = "void" }
                            else
                                raise (Failure ("Second argument passed to function 'imsave' must be of"
                                ^ " string type, called from function: '" ^ fname ^ "'"))
                        else
                            raise (Failure ("First argument passed to function 'imsave' must be of"
                            ^ " image type, called from function: '" ^ fname ^ "'"))
      | Call("imshow", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'imshow'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "image" then 
                            { varname = "imshow" ; vartype = "void" }
                        else
                            raise (Failure ("Argument passed to function 'imshow' must be of image type" 
                                ^ ", called from function: '" ^ fname ^ "'"))
      | Call("getwidth", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'getwidth'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "image" || param.vartype = "kernel" then 
                            { varname = "getwidth" ; vartype = "int" }
                        else
                            raise (Failure ("Argument passed to function 'getwidth' must be of either "
                            ^ "image or kernel type, called from function: '" ^ fname ^ "'"))
      | Call("getheight", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'getheight'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "image" || param.vartype = "kernel" then 
                            { varname = "getheight" ; vartype = "int" }
                        else
                            raise (Failure ("Argument passed to function 'getheight' must be of either "
                            ^ "image or kernel type, called from function: '" ^ fname ^ "'"))
      | Call("getlength", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'getlength'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "string" then 
                            { varname = "getlength" ; vartype = "int" }
                        else
                            raise (Failure ("Argument passed to function 'getlength' must be of "
                            ^ "string type, called from function: '" ^ fname ^ "'"))
      | Call("gettype", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("Wrong number of arguments passed to function 'gettype'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "image" then 
                            { varname = "gettype" ; vartype = "string" }
                        else
                            raise (Failure ("Argument passed to function 'gettype' must be of "
                            ^ "image type, called from function: '" ^ fname ^ "'"))
      | Call("mod", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 2 then 
                    raise (Failure ("Wrong number of arguments passed to function 'mod'" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param1 = List.nth actuals 1 in
                    let param2 = List.nth actuals 0 in
                        if param1.vartype = "int" && param2.vartype = "int" then
                            { varname = "mod" ; vartype = "int" }
                        else if param1.vartype = "float" && param2.vartype = "float" then
                            { varname = "mod" ; vartype = "float" }
                        else
                            raise (Failure ("Arguments passed to function 'mod' must be all of"
                            ^ " either int or float type,"
                            ^ " but here got '" ^ param1.vartype ^ "' and '" 
	                        ^ param2.vartype ^ "' called from function: '" ^ fname ^ "'"))
      | Call("sqrt", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 1 then 
                    raise (Failure ("function 'sqrt' takes just one parameter" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    let param = List.hd actuals in
                        if param.vartype = "float" || param.vartype = "int" then 
                            { varname = "sqrt" ; vartype = "float" }
                        else
                            raise (Failure ("Argument passed to function 'sqrt' can only be " 
                                ^ "either float or int type, called from function: '" ^ fname ^ "'"))
      | Call("read", actuals) ->
            let actuals = List.fold_left 
                (fun actuals actual ->
	            let v = string_of_expr globals locals fdecl.fname actual in
	                v :: actuals) [] actuals
                in
                if (List.length actuals) <> 0 then 
                    raise (Failure ("function 'read' does not take any parameter" 
                        ^ ", called from function: '" ^ fname ^ "'"))
                else
                    { varname = "read" ; vartype = "string" }
      | Call(f, actuals) ->
            let fdecl =
                if f = "main" then
                    raise (Failure ("main function cannot be called by " ^
	                                         "function: '" ^ fname ^ "'"))
                else
	            try NameMap.find f func_decls
	                with Not_found -> raise (Failure ("Undefined function: '" 
	                    ^ f ^ "' in function call: '" ^ fname ^ "'"))
	            in
	            if f <> fname then
    	            let actuals = List.fold_left 
    	                (fun actuals actual ->
    		            let v = string_of_expr globals locals fdecl.fname actual in
    		                v :: actuals) [] actuals
    	                in
                            string_of_fdecl globals fdecl (List.rev actuals);
                            { varname = fdecl.retname ; vartype = fdecl.rettype }
                else
                    let actuals = List.fold_left 
    	                (fun actuals actual ->
    		            let v = string_of_expr globals locals fdecl.fname actual in
    		                v :: actuals) [] actuals
    	                in
                (try List.iter2 (fun formal actual -> 
                    if formal.vartype <> actual.vartype then
                        raise (Failure ("Type mismatch in argument passing between: '" 
                        ^ formal.varname ^ "', type: " ^ formal.vartype ^ " and '" 
                        ^ actual.varname ^ "', type: " ^ actual.vartype 
                        ^ " in function: '" ^ fdecl.fname ^ "'")))
                       fdecl.formals (List.rev actuals)
                (* Check invalid number of arguments *)    
                with Invalid_argument(_) ->
                    raise (Failure ("Wrong number of arguments passed to function: '" 
                        ^ fdecl.fname ^ "' from function: '" ^ fname ^ "'")));
                { varname = fdecl.retname ; vartype = fdecl.rettype }
    in
    
    let check_for_expr v l globals locals fname =
        if v.vartype <> "int" or v.vartype <> "float" or v.vartype <> "double" then
            let varList = List.map (string_of_expr globals locals fname) l
	        in
    	    List.iter (fun var -> if var.vartype <> v.vartype then  
    	            raise (Failure ("For loop type mismatch between '" ^ v.varname 
    	            ^ "', type: " ^ v.vartype ^ " and '" ^ var.varname 
    	            ^ "', type: " ^ var.vartype ^ " in function: '" ^ fname ^ "'"))) varList;
	    else
	        raise (Failure ("Only int/float is allowed in the for loop. Id '" 
	            ^ v.varname ^ "', type: " ^ v.vartype ^ " found in function: '" ^ fname ^ "'"))
    in	        
    
    let rec string_of_for_expr globals locals fname = function
        Assigna(v, e1, e2) -> 
        if NameMap.mem v locals then
            let var = NameMap.find v locals in
            check_for_expr var [e1;e2;] globals locals fname
        else if NameMap.mem v globals then
	        let var = NameMap.find v globals in
	        check_for_expr var [e1;e2;] globals locals fname
	    else raise (Failure ("Undeclared identifier in for loop: '" ^ v 
	        ^ "' in function: '" ^ fname ^ "'"))
    |   Assignb(v, e1, e2, e3) ->
        if NameMap.mem v locals then
            let var = NameMap.find v locals in
            check_for_expr var [e1;e2;e3;] globals locals fname
	    else if NameMap.mem v globals then
	        let var = NameMap.find v globals in
	        check_for_expr var [e1;e2;e3] globals locals fname
        else raise (Failure ("Undeclared identifier in for loop: '" ^ v
            ^ "' in function: '" ^ fname ^ "'"))
    in
    let checkDimension kernel fname l = 
        let rec loop l = 
            if (List.length l) = 1 then 
                List.length (List.hd l)
            else
                let c1 = List.length (List.hd l)
                and c2 = loop (List.tl l) in
                if (c1 <> c2) then
                    raise (Failure ("Invalid kernel initialization: number of columns"
                    ^ " in all rows must be same. kernel: '" ^ kernel ^ "' ;"
                    ^"function: '" ^ fname ^ "'"))
                else c1
            in loop l
    in
    let checkKernelValues globals locals kernel fname l = 
        let flatL = List.concat l in
            List.iter
            (fun exp ->
            let v = string_of_expr globals locals fname exp in
                if v.vartype <> "float" then
                    raise (Failure ("Invalid kernel value: all values passed to "
                        ^ " kernel must be float type. kernel: '" ^ kernel ^ "' ;"
                    ^"function: '" ^ fname ^ "'"))) flatL
    in
    let blackhole e = () 
    in
    let rec string_of_stmt globals locals fname = function
        Block(stmts) ->
            List.iter (string_of_stmt globals locals fname) stmts
      | Expr(expr) -> let e = string_of_expr globals locals fname expr in blackhole e
      | If(e, s, Block([])) -> let v = string_of_expr globals locals fname e in 
            if v.vartype <> "boolean" then
                raise (Failure ("if expression must be evaluated to boolean, "
                    ^ "but here found with '" ^ v.vartype ^ "' ;"
                    ^ "function: '" ^ fname ^ "'"))
            else
                string_of_stmt globals locals fname s; 
      | If(e, s1, s2) ->  let v = string_of_expr globals locals fname e in 
            if v.vartype <> "boolean" then
                raise (Failure ("if expression must be evaluated to boolean, "
                    ^ "but here found with '" ^ v.vartype ^ "' ;"
                    ^ "function: '" ^ fname ^ "'"))
            else
                string_of_stmt globals locals fname s1;
                string_of_stmt globals locals fname s2;
      | Ifelseif(e, s, eif_list, Block([])) -> 
            let v = string_of_expr globals locals fname e in 
            if v.vartype <> "boolean" then
                raise (Failure ("if expression must be evaluated to boolean, "
                    ^ "but here found with '" ^ v.vartype ^ "' ;"
                    ^ "function: '" ^ fname ^ "'"))
            else
                let rec string_of_elseif eif = 
                      let en = string_of_expr globals locals fname eif.elseif_expr in
                        if en.vartype <> "boolean" then
                            raise (Failure ("if expression must be evaluated to boolean, "
                                ^ "but here found with '" ^ v.vartype ^ "' ;"
                                ^ "function: '" ^ fname ^ "'"))
                        else
                            string_of_stmt globals locals fname eif.elseif_stmt;
                in
                    string_of_stmt globals locals fname s;
                    List.iter string_of_elseif eif_list;
      | Ifelseif(e, s1, eif_list, s2) -> 
            let v = string_of_expr globals locals fname e in 
            if v.vartype <> "boolean" then
                raise (Failure ("if expression must be evaluated to boolean, "
                    ^ "but here found with '" ^ v.vartype ^ "' ;"
                    ^ "function: '" ^ fname ^ "'"))
            else
                let rec string_of_elseif eif = 
                    let en = string_of_expr globals locals fname eif.elseif_expr in
                        if en.vartype <> "boolean" then
                            raise (Failure ("if expression must be evaluated to boolean, "
                                ^ "but here found with '" ^ v.vartype ^ "' ;"
                                ^ "function: '" ^ fname ^ "'"))
                        else
                            string_of_stmt globals locals fname eif.elseif_stmt;
                in
                    string_of_stmt globals locals fname s1;
                    List.iter string_of_elseif eif_list;
                    string_of_stmt globals locals fname s2;
      | For(e, s) ->
                string_of_for_expr globals locals fname e;
                string_of_stmt globals locals fname s
      | While(e, s) -> let v = string_of_expr globals locals fname e 
            in 
            if v.vartype <> "boolean" then
                raise (Failure ("while expression must be evaluated to boolean, "
                    ^ "but here found with '" ^ v.vartype ^ "' ;"
                    ^ "function: '" ^ fname ^ "'"))
            else
                string_of_stmt globals locals fname s; 
      | KernelInit(id, valuesList) ->
            if NameMap.mem id locals then
                let varId = NameMap.find id locals in
                if varId.vartype <> "kernel" then
    	            raise (Failure ("Expecting kernel type but got '" ^ varId.vartype ^
    	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
    	        else 
    	            let col = checkDimension id fname valuesList in
    	                blackhole col;
    	                checkKernelValues globals locals id fname valuesList
    	    else if NameMap.mem id globals then
                let varId = NameMap.find id globals in
                if varId.vartype <> "kernel" then
    	            raise (Failure ("Expecting kernel type but got '" ^ varId.vartype ^
    	                "' for variable: '" ^ varId.varname ^ " in function: '" ^ fname ^ "'")) 
    	        else
    	            let col = checkDimension id fname valuesList in
    	                blackhole col;
    	                checkKernelValues globals locals id fname valuesList
    	    else raise (Failure ("Undeclared identifier found in function: '" ^ fname ^ "'"))
  
    in
    let locals = 
        try List.fold_left2
            (fun locals formal actual -> 
                if formal.vartype = actual.vartype then
                    if NameMap.mem formal.varname locals then
                        raise (Failure ("Formal variable: '" ^ formal.varname ^ 
                            "' has already been defined in the function: '" ^ 
                            fdecl.fname ^ "' parameter"))
                    else NameMap.add formal.varname formal locals
                else raise (Failure ("Type mismatch in argument passing between: '" 
    	                ^ formal.varname ^ "', type: " ^ formal.vartype ^ " and '" 
    	                ^ actual.varname ^ "', type: " ^ actual.vartype 
    	                ^ " in function: '" ^ fdecl.fname ^ "'")))
	            NameMap.empty fdecl.formals actuals
	    (* Check invalid number of arguments *)    
        with Invalid_argument(_) ->
	        raise (Failure ("Wrong number of arguments passed to function: '" 
	            ^ fdecl.fname ^ "'"))    
        in      
        let locals = List.fold_left
	        (fun locals local -> 
                if NameMap.mem local.varname locals then
                        raise (Failure ("Local variable: '" ^ local.varname ^ 
                            "' has already been defined in the function: '" ^ 
                            fdecl.fname ^ "'"))
                else NameMap.add local.varname local locals) locals fdecl.locals
            in  
            if fdecl.retname <> "" then
                if NameMap.mem fdecl.retname locals then 
                    raise (Failure ("Return variable: '" ^ fdecl.retname ^ 
                        "' has already been defined in the function: '" ^ 
                        fdecl.fname ^ "'"))
                else
                    let ret = { varname = fdecl.retname; vartype = fdecl.rettype } in
                    let locals = NameMap.add ret.varname ret locals in
                    List.iter (string_of_stmt globals locals fdecl.fname) fdecl.body
            else
                List.iter (string_of_stmt globals locals fdecl.fname) fdecl.body
    in
    let globals = List.fold_left
        (fun globals vdecl -> 
            if NameMap.mem vdecl.varname globals then
                raise (Failure ("Global variable : '" ^ vdecl.varname ^ 
                            "' has already been defined."))
            else
                NameMap.add vdecl.varname vdecl globals) NameMap.empty vars
    in 
        try
            string_of_fdecl globals (NameMap.find "main" func_decls) []
        with Not_found -> 
            raise (Failure ("There is no main() function"))        

